Attribute VB_Name = "mLBSbCls"
' Listbox Subclassing: Ported the C code found on MSDN at
'      SDK Documentation/PlatformSDK/User Interace services/controls/listbox
' and Knowledgebase Article ID: Q74792
'
' By T. Ravi Kiran.
' E-Mail: t_ravi_kiran@rocketmail.com
' You can also send me messages at CodeGuru site: http://www.codeguru.com/vb
'
'
' List box sub classing: To achieve Non-selectable items in a ListView
'
' What This piece of code does:
'   1. Display some predefined ListItems in Gray (inactive) color.
'   2. Dis allow user selection of these items by mouse (click)
'   3. When scrolled/browsed by Arrow keys, jump over these non-active items
'
'
' How to Use:
'   Step 1: Take a form and put a list box.
'   Step 2: Set the Listbox and Parent control's pointers of this module to them
'   Step 3: Call SetHooks to setup the subclass procedures.
'   Step 4: Run.
'   Step 5: REMEMBER to unhook by callikg UnSetHooks in the end.
'
' Notes: You will find a lot of commented out code. That is because i am
' planning for future upgrades
'
' Advice: If you are debuggin this code, DONT put break points in subclassed
' Window procedures. VB IDE could become unstable. Always use Debug.Print
' to print any variable's value. Save before running, PLEASE..
'
Option Explicit

'
' Win API Rect structure
'
Private Type RECT
    Left As Long
    Top As Long
    Right As Long
    Bottom As Long
End Type
' The Draw item structure for Owener Draw items of basic control types
Type DRAWITEMSTRUCT
    CtlType As Long
    CtlID As Long
    itemID As Long
    itemAction As Long
    itemState As Long
    hwndItem As Long
    hdc As Long
    rcItem As RECT
    itemData As Long
End Type
'
' Owner draw control types
'
'Const ODT_MENU = 1
Const ODT_LISTBOX = 2
'Const ODT_COMBOBOX = 3
'Const ODT_BUTTON = 4
'Const ODT_STATIC = 5

' Standard windows style
Const WS_VSCROLL = &H200000
Const WS_HSCROLL = &H100000
Const WS_BORDER = &H800000

Public Const GWL_STYLE = (-16)
Public Const GWL_EXSTYLE = (-20)

' List Box Styles:
'Public Const LBS_NOTIFY = &H1
'Public Const LBS_SORT = &H2
'Public Const LBS_NOREDRAW = &H4
Public Const LBS_MULTIPLESEL = &H8
Public Const LBS_OWNERDRAWFIXED = &H10
'Public Const LBS_OWNERDRAWVARIABLE = &H20
'Public Const LBS_HASSTRINGS = &H40
'Public Const LBS_USETABSTOPS = &H80
'Public Const LBS_NOINTEGRALHEIGHT = &H100
'Public Const LBS_MULTICOLUMN = &H200
'Public Const LBS_WANTKEYBOARDINPUT = &H400
'Public Const LBS_EXTENDEDSEL = &H800
'Public Const LBS_DISABLENOSCROLL = &H1000
'Public Const LBS_NODATA = &H2000
'Public Const LBS_NOSEL = &H4000
'Public Const LBS_STANDARD = (LBS_NOTIFY Or LBS_SORT Or WS_VSCROLL Or WS_BORDER)

'List box Notification messages:
Const LBN_SELCHANGE = 1
Const LBN_DBLCLK = 2
Const LBN_SELCANCEL = 3
Const LBN_SETFOCUS = 4
Const LBN_KILLFOCUS = 5

' List box Messages: This is the complete set. In this example only few
' are used.
'Const LB_ADDSTRING = &H180
'Const LB_INSERTSTRING = &H181
'Const LB_DELETESTRING = &H182
'Const LB_SELITEMRANGEEX = &H183
'Const LB_RESETCONTENT = &H184
'Const LB_SETSEL = &H185
Const LB_SETCURSEL = &H186
'Const LB_GETSEL = &H187
Const LB_GETCURSEL = &H188
Const LB_GETTEXT = &H189
Const LB_GETTEXTLEN = &H18A
'Const LB_GETCOUNT = &H18B
'Const LB_SELECTSTRING = &H18C
'Const LB_DIR = &H18D
'Const LB_GETTOPINDEX = &H18E
'Const LB_FINDSTRING = &H18F
'Const LB_GETSELCOUNT = &H190
'Const LB_GETSELITEMS = &H191
'Const LB_SETTABSTOPS = &H192
'Const LB_GETHORIZONTALEXTENT = &H193
'Const LB_SETHORIZONTALEXTENT = &H194
'Const LB_SETCOLUMNWIDTH = &H195
'Const LB_ADDFILE = &H196
'Const LB_SETTOPINDEX = &H197
'Const LB_GETITEMRECT = &H198
'Const LB_GETITEMDATA = &H199
'Const LB_SETITEMDATA = &H19A
'Const LB_SELITEMRANGE = &H19B
'Const LB_SETANCHORINDEX = &H19C
'Const LB_GETANCHORINDEX = &H19D
Const LB_SETCARETINDEX = &H19E
Const LB_GETCARETINDEX = &H19F
'Const LB_SETITEMHEIGHT = &H1A0
'Const LB_GETITEMHEIGHT = &H1A1
'Const LB_FINDSTRINGEXACT = &H1A2
'Const LB_SETLOCALE = &H1A5
'Const LB_GETLOCALE = &H1A6
'Const LB_SETCOUNT = &H1A7
'Const LB_INITSTORAGE = &H1A8
Const LB_ITEMFROMPOINT = &H1A9
'Const LB_MSGMAX = &H1B0


' The Windows Standard colors constants:- To be used with GetSysColor call
Const COLOR_WINDOW = 5
Const COLOR_WINDOWTEXT = 8
Const COLOR_HIGHLIGHT = 13
Const COLOR_HIGHLIGHTTEXT = 14
Const COLOR_GRAYTEXT = 17


'
' Notify Message
'
Private Const WM_NOTIFY& = &H4E
Public Const WM_COMMAND = &H111
Const WM_DRAWITEM = &H2B

' Mouse Constants:-
'Const WM_MOUSEFIRST = &H200
Const WM_MOUSEMOVE = &H200
Const WM_LBUTTONDOWN = &H201
'Const WM_LBUTTONUP = &H202
Const WM_LBUTTONDBLCLK = &H203
'Const WM_RBUTTONDOWN = &H204
'Const WM_RBUTTONUP = &H205
'Const WM_RBUTTONDBLCLK = &H206
'Const WM_MBUTTONDOWN = &H207
'Const WM_MBUTTONUP = &H208
'Const WM_MBUTTONDBLCLK = &H209

' Windows messages related to keyboard:
Const WM_KEYDOWN = &H100
Const WM_CHAR = &H102

Const WM_GETFONT = &H31
Const WM_PAINT = &HF
Const WM_ERASEBKGND = &H14


' ** Constants relating to the selected status of list item:
Const ODA_FOCUS = &H4       ' This is for the List box of Combo box!
Const ODS_FOCUS = &H10
Const ODS_SELECTED = &H1

'
' Win API Declarations
'
Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (pDest As Any, pSource As Any, ByVal dwLength As Long)
Public Declare Function SendMessage Lib "user32" Alias "SendMessageA" (ByVal hwnd As Long, ByVal wMsg As Long, ByVal wParam As Long, lParam As Any) As Long

Declare Function DrawText Lib "user32" Alias "DrawTextA" (ByVal hdc As Long, ByVal lpStr As String, ByVal nCount As Long, lpRect As RECT, ByVal wFormat As Long) As Long
'Declare Function DrawTextEx Lib "user32" Alias "DrawTextExA" (ByVal hdc As Long, ByVal lpsz As String, ByVal n As Long, lpRect As RECT, ByVal un As Long, lpDrawTextParams As DRAWTEXTPARAMS) As Long
Declare Function SetTextAlign Lib "gdi32" (ByVal hdc As Long, ByVal wFlags As Long) As Long
Declare Function SetTextColor Lib "gdi32" (ByVal hdc As Long, ByVal crColor As Long) As Long
Declare Function SetBkColor Lib "gdi32" (ByVal hdc As Long, ByVal crColor As Long) As Long
Declare Function InvalidateRect Lib "user32" (ByVal hwnd As Long, lpRect As RECT, ByVal bErase As Long) As Long
Declare Function BitBlt Lib "gdi32" (ByVal hDestDC As Long, ByVal X As Long, _
    ByVal Y As Long, ByVal nWidth As Long, ByVal nHeight As Long, _
    ByVal hSrcDC As Long, ByVal xSrc As Long, ByVal ySrc As Long, _
    ByVal dwRop As Long) As Long

Declare Function DrawFocusRect Lib "user32" (ByVal hdc As Long, lpRect As RECT) As Long
Declare Function CreateSolidBrush Lib "gdi32" (ByVal crColor As Long) As Long
Declare Function GrayString Lib "user32" Alias "GrayStringA" (ByVal hdc As Long, ByVal hBrush As Long, ByVal lpOutputFunc As Long, _
        ByVal lpData As String, ByVal nCount As Long, _
        ByVal X As Long, ByVal Y As Long, ByVal nWidth As Long, ByVal nHeight As Long) As Long
'Declare Function GrayString Lib "user32" Alias "GrayStringA" (ByVal hdc As Long, ByVal hBrush As Long, ByVal lpOutputFunc As Long, _
        ByVal lpData As Long, ByVal nCount As Long, _
        ByVal x As Long, ByVal y As Long, ByVal nWidth As Long, ByVal nHeight As Long) As Long
Declare Function DeleteObject Lib "gdi32" (ByVal hObject As Long) As Long
Declare Function GetSysColor Lib "user32" (ByVal nIndex As Long) As Long
Declare Function FillRect Lib "user32" (ByVal hdc As Long, lpRect As RECT, ByVal hBrush As Long) As Long
Declare Function TextOut Lib "gdi32" Alias "TextOutA" (ByVal hdc As Long, ByVal X As Long, ByVal Y As Long, ByVal lpString As String, ByVal nCount As Long) As Long
' I couldn't get the SendMessage(LB_GETTEXT) to work with VB's string
' variable. So i modified the definition to handle byte-arrays for both
' SendMessage and TextOut API calls. So if you are using this with in similar
' definitions please take note of.
' This defn works with both Eng & Japanese strings. So i guess it is ok.
Declare Function TextOutBStr Lib "gdi32" Alias "TextOutA" (ByVal hdc As Long, ByVal X As Long, ByVal Y As Long, lpString As Any, ByVal nCount As Long) As Long


Public Const GW_OWNER = 4
Public Const GWL_WNDPROC = (-4)
Public Declare Function GetWindowLong Lib "user32" Alias "GetWindowLongA" (ByVal hwnd As Long, ByVal nIndex As Long) As Long
Declare Function SetWindowLong Lib "user32" Alias "SetWindowLongA" _
    (ByVal hwnd As Long, ByVal nIndex As Long, ByVal dwNewLong As Long) As Long
Declare Function CallWindowProc Lib "user32" Alias "CallWindowProcA" _
    (ByVal lpPrevWndFunc As Long, ByVal hwnd As Long, ByVal Msg As Long, _
    ByVal wParam As Long, ByVal lParam As Long) As Long



'' This module related constants
Const TwoPower16 = 2 ^ 16       '!!. I need this for lparam to/from conv.

Private lpListBox As ListBox
Private lpParent As Form     '<-- Change this in case the parent is not Form
            ' Any VB container control will do. Ex Frame, Picturebox, Form

Public lpHBitmap(3) As StdPicture ' for the 3 states of Multi-selection LB item

Public oldWndProc As Long

Private LBProc1 As Long
Private m_Hooked_LBhWnd As Long
Private m_LBHwnd As Long
Private m_Hooked_hWnd As Long
Private m_bCustomDraw As Boolean

Public Function SetListBox(lpLB As ListBox) As Boolean
    On Error GoTo errset
    Set lpListBox = lpLB
    m_LBHwnd = lpListBox.hwnd
    SetListBox = True
    Exit Function
errset:
End Function
Public Function SetLBParent(lpLBBoss As Form) As Boolean
    On Error GoTo errset
    Set lpParent = lpLBBoss
    SetLBParent = True
    Exit Function
errset:
End Function


' This is for subclassing the parent, so that we can listen to list box's
' Notification messages
Private Function LBSubcls_WndProc_V3(ByVal hwnd As Long, ByVal Msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    Dim iHw As Integer, iLW As Integer
    Dim lCurind As Long
    Select Case Msg
        Case WM_COMMAND
            If lParam = m_LBHwnd Then
                LongInt2Int wParam, iHw, iLW
                Select Case (iHw)
                Case LBN_SELCHANGE
'                    lCurind = SendMessage(lParam, LB_GETCURSEL, 0, ByVal 0&)
'                    Debug.Print " LBN selchange with for"; Hex$(lCurind)
'                    'If IndexAllowed(lCurind) = False Then
'                    ' You might need a fn like above: For my case
'                    If (lCurind Mod 3) = 0 Then
'                    lCurind = SendMessage(lParam, LB_SETCURSEL, lCurind + 1,ByVal 0&)
'                    End If
                    'Debug.Print " sendmessage returned:"; Hex$(lCurind)
                Case LBN_SELCANCEL
                    lCurind = SendMessage(lParam, LB_GETCURSEL, 0, ByVal 0&)
                    Debug.Print " lbnselcancel for:"; Hex$(lCurind)
                Case LBN_DBLCLK
                'Case LBN_KILLFOCUS
                'Case LBN_SETFOCUS
                End Select
            End If
        Case WM_DRAWITEM
            If LB_Drawitem(lParam) = 0 Then
                ' we have handled the painting. So dont pass on to
                ' default window procedure
                LBSubcls_WndProc_V3 = 0
                Exit Function
            End If
        Case Else
    End Select
    '
    ' ** for other things do default behaviour:
    '
    LBSubcls_WndProc_V3 = CallWindowProc(oldWndProc, hwnd, Msg, wParam, lParam)

End Function

Public Sub SetHooks()
    If lpListBox Is Nothing Then
        MsgBox "List box pointer not set. Wrong sequence of calls"
        Exit Sub
    End If
    If lpParent Is Nothing Then
        MsgBox "parent pointer for the LB not set. Wrong sequence of calls"
        Exit Sub
    End If
    SetHookLB lpListBox.hwnd, True
    SetHookParent lpParent.hwnd, True
End Sub
Public Sub UnSetHooks()
    SetHookLB vbNull, False
    SetHookParent vbNull, False
End Sub

Public Sub SetHookParent(ByVal hwnd As Long, b As Boolean)
    If b Then
        oldWndProc = SetWindowLong(hwnd, GWL_WNDPROC, AddressOf LBSubcls_WndProc_V3)
        m_Hooked_hWnd = hwnd
        m_bCustomDraw = True
    Else
        Call SetWindowLong(m_Hooked_hWnd, GWL_WNDPROC, oldWndProc)
        m_Hooked_hWnd = 0
    End If
End Sub

Public Sub SetHookLB(ByVal LBhWnd As Long, b As Boolean)
    If b Then
        LBProc1 = SetWindowLong(LBhWnd, GWL_WNDPROC, AddressOf LBSubcls_WndProc_V4)
        m_Hooked_LBhWnd = LBhWnd
    Else
        Call SetWindowLong(m_Hooked_LBhWnd, GWL_WNDPROC, LBProc1)
        m_Hooked_LBhWnd = 0
    End If
End Sub

' Item Draw Notification Event handler for List Box:
Private Function LB_Drawitem(ByVal lParam As Long) As Integer
    ' The input is actually a pointer to a structure of type
    ' DrawItemStruct
    Dim drawstruct As DRAWITEMSTRUCT
    Dim szBuf(256) As Byte
    ' The 'C's way of pointer casting has to implemented in VB by
    ' copying.
    CopyMemory drawstruct, ByVal lParam, Len(drawstruct)

    ' Lets not allow every 3rd item to be selected.
    Dim i As Integer
    Dim hbrGray As Long, hbrback As Long, szListStr As String ' * 256
    Dim crback As Long, crtext As Long, lbuflen As Long
    Select Case (drawstruct.CtlType)
    Case ODT_LISTBOX:
        ' Get the String:
        lbuflen = SendMessage(drawstruct.hwndItem, _
                          LB_GETTEXTLEN, _
                          drawstruct.itemID, ByVal 0&)
        ' ideally you should redim : Redim szBuf(lbuflen+2)
        '
        ' I couldn't get the SendMessage(LB_GETTEXT) to work with VB's string
        ' variable. So i modified the definition to handle byte-arrays for both
        ' SendMessage and TextOut API calls - see defs section on top.
        lbuflen = SendMessage(drawstruct.hwndItem, _
                          LB_GETTEXT, _
                          drawstruct.itemID, szBuf(0))

        i = drawstruct.itemID
        '' ** Change this condition to suite your requirement***
        '' Only Ex: Every 3rd element is not allowed.
        If i Mod 3 = 0 Then
'            hbrGray = CreateSolidBrush(GetSysColor(COLOR_GRAYTEXT))
'            GrayString drawstruct.hdc, hbrGray, ByVal 0&, szListStr, Len(szListStr), _
'                      drawstruct.rcItem.Left, drawstruct.rcItem.Top, _
'                      0, 0
'            DeleteObject hbrGray
            crback = RGB(180, 180, 180)
            crtext = RGB(60, 60, 60)
        Else
            If (drawstruct.itemState And ODS_SELECTED) = ODS_SELECTED Then
                '/* Set background and text colors for selected item */
                crback = GetSysColor(COLOR_HIGHLIGHT)
                crtext = GetSysColor(COLOR_HIGHLIGHTTEXT)
                ' Or you could use other colors and be un-Windows like!
                'crback = vbGreen
                'crtext = vbWhite
            ElseIf (drawstruct.itemState And ODS_FOCUS) = ODS_FOCUS Then
                ' For Multi selection listbox, this state exists - Curnt focus
                ' choose a seperate color if we want!
                crback = GetSysColor(COLOR_WINDOW)
                crtext = vbRed
            Else
                '/* Set background and text colors for unselected item */
                ' Default windows colors:-
                'crback = GetSysColor(COLOR_WINDOW)
                'crtext = GetSysColor(COLOR_WINDOWTEXT)
                ' Some colors:-
                'crback = vbBlue
                'crtext = vbCyan
                'Since VB supports Design time setting of colors:
                crback = lpListBox.BackColor
                crtext = lpListBox.ForeColor
            End If
            ' Some un-Windows effects:-
            If (drawstruct.itemState And ODS_FOCUS) = ODS_FOCUS Then
                crtext = vbRed
            End If
        End If
        '// Fill item rectangle with background color
        hbrback = CreateSolidBrush(crback)
        FillRect drawstruct.hdc, drawstruct.rcItem, hbrback
        DeleteObject hbrback

        '// Set current background and text colors
        SetBkColor drawstruct.hdc, crback
        SetTextColor drawstruct.hdc, crtext

        '// TextOut uses current background and text colors
'            TextOut drawstruct.hdc, _
'                      drawstruct.rcItem.Left, _
'                      drawstruct.rcItem.Top, _
'                      szListStr, Len(szListStr)
        TextOutBStr drawstruct.hdc, _
                  drawstruct.rcItem.Left, _
                  drawstruct.rcItem.Top, _
                   szBuf(0), lbuflen
        '/* If enabled item has the input focus, call
        'DrawFocusRect to set or clear the focus
        'rectangle */
        If (drawstruct.itemState And ODS_FOCUS) Then
            DrawFocusRect drawstruct.hdc, drawstruct.rcItem
        End If

        ' If we handled the painting we will pass 0, else 1
        LB_Drawitem = 0
''''    ' Other Possible cases:-
''''    Case ODT_COMBOBOX:
''''    Case ODT_BUTTON
''''    Case ODT_MENU
''''    Case ODT_STATIC
        Case Else ' IMPORTANT!- We are not handling other guys!
            LB_Drawitem = 1
    End Select
End Function

' This function is damm performance hit. Please tell me if you
' can find a better one to get the higher word from longword
Public Function LongInt2Int(ByVal lLongInt As Long, _
    ByRef iHiWord As Integer, ByRef iLowWord As Integer) As Boolean
    Dim tmpHW As Integer, tmpLW As Integer
    CopyMemory tmpLW, lLongInt, Len(tmpLW)
    tmpHW = (lLongInt / TwoPower16)
    iHiWord = tmpHW
    iLowWord = tmpLW
End Function
Public Function MakeLParam(ByVal iHiWord As Integer, ByVal iLowWord As Integer) As Long
    MakeLParam = (iHiWord * TwoPower16) + iLowWord
End Function

' This is to sub class the List box itself
' Process Mouse click & DBL click on items and reject clicks on not allowed ones
' Process key down messages and jump over not allowed ones.

Private Function LBSubcls_WndProc_V4(ByVal hwnd As Long, ByVal Msg As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
    Dim iHw As Integer, iLW As Integer
    Dim lCurind As Long
    Select Case Msg
        Case WM_LBUTTONDOWN, WM_LBUTTONDBLCLK
            LongInt2Int lParam, iHw, iLW
            Debug.Print " Mouse down at("; iHw; ","; iLW; ")";
            lCurind = SendMessage(hwnd, LB_ITEMFROMPOINT, ByVal 0, ByVal lParam)
            Debug.Print "Index of btn down:"; Hex$(lCurind)
            If (lCurind Mod 3) = 0 Then
                LBSubcls_WndProc_V4 = 1
                ' by doing this we will be eating away the click on
                ' unwanted items.
                Exit Function
            End If
        Case WM_KEYDOWN
            LongInt2Int wParam, iHw, iLW
            Select Case (iLW)
            Case vbKeyDown
                lCurind = SendMessage(hwnd, LB_GETCARETINDEX, 0, ByVal 0&)
                Debug.Print " Keydown with Itemid :"; Hex$(lCurind)
                If ((lCurind + 1) Mod 3) = 0 Then
                    lCurind = SendMessage(hwnd, LB_SETCARETINDEX, lCurind + 1, ByVal 0&)
                End If
                lCurind = SendMessage(hwnd, LB_GETCURSEL, 0, ByVal 0&)
                Debug.Print " Keydown with Itemid :"; Hex$(lCurind)
                If ((lCurind + 1) Mod 3) = 0 Then
                    lCurind = SendMessage(hwnd, LB_SETCURSEL, lCurind + 1, ByVal 0&)
                End If
                ' Notice that we are still letting the Focus rect be
                ' drawn by default window proc. We are just changing the
                ' Caret item /focus item.
            Case vbKeyUp
                lCurind = SendMessage(hwnd, LB_GETCARETINDEX, 0, ByVal 0&)
                Debug.Print " KeyUp with Itemid :"; Hex$(lCurind)
                If ((lCurind - 1) Mod 3) = 0 Then
                    lCurind = SendMessage(hwnd, LB_SETCARETINDEX, lCurind - 1, ByVal 0&)
                End If
                lCurind = SendMessage(hwnd, LB_GETCURSEL, 0, ByVal 0&)
                Debug.Print " KeyUp with Itemid :"; Hex$(lCurind)
                If ((lCurind - 1) Mod 3) = 0 Then
                    lCurind = SendMessage(hwnd, LB_SETCURSEL, lCurind - 1, ByVal 0&)
                End If
            ' These two are also to be handled:-
            Case vbKeyPageUp
            Case vbKeyPageDown
            End Select
        Case Else
    End Select
    '
    ' ********* AS OF NOW WE ARE ONLY DOING DEFAULT BEHAVIOUR *********
    '
    LBSubcls_WndProc_V4 = CallWindowProc(LBProc1, hwnd, Msg, wParam, lParam)

End Function

